# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import frost_sta_client.model
from . import entity
from . import location
from . import thing

from frost_sta_client.dao.historical_location import HistoricalLocationDao

from frost_sta_client import utils
from .ext import entity_list
from .ext import entity_type


class HistoricalLocation(entity.Entity):

    def __init__(self,
                 locations=None,
                 time=None,
                 thing=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.locations = locations
        self.time = time
        self.thing = thing

    def __new__(cls, *args, **kwargs):
        new_h_loc = super().__new__(cls)
        attributes = {'_id': None, '_locations': None, '_time': None, '_thing': None, '_self_link': None,
                      '_service': None}
        for key, value in attributes.items():
            new_h_loc.__dict__[key] = value
        return new_h_loc

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, value):
        self._time = utils.check_datetime(value, 'time')

    @property
    def locations(self):
        return self._locations

    @locations.setter
    def locations(self, values):
        if values is None:
            self._locations = None
            return
        if isinstance(values, list) and all(isinstance(loc, location.Location) for loc in values):
            entity_class = entity_type.EntityTypes['Location']['class']
            self._locations = entity_list.EntityList(entity_class=entity_class, entities=values)
            return
        if isinstance(values, entity_list.EntityList) and\
                all((isinstance(loc, location.Location)) for loc in values.entities):
            self._locations = values
            return
        raise ValueError('locations should be a list of locations!')

    @property
    def thing(self):
        return self._thing

    @thing.setter
    def thing(self, value):
        if value is None or isinstance(value, thing.Thing):
            self._thing = value
            return
        raise ValueError('thing should be of type Thing!')

    def ensure_service_on_children(self, service):
        if self.locations is not None:
            self.locations.set_service(service)
        if self.thing is not None:
            self.thing.set_service(service)

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, type(self)):
            return False
        if id(self) == id(other):
            return True
        if self.time != other.time:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        data = super().__getstate__()
        if data.get('@iot.id', None) is not None:
            return data
        if self.time is not None:
            data['Time'] = utils.parse_datetime(self.time)
        if self.thing is not None:
            data['Thing'] = self.thing
        if self.locations is not None and len(self.locations.entities) > 0:
            data['Locations'] = self.locations.__gestate__()
        return data

    def __setstate__(self, state):
        super().__setstate__(state)
        self.time = state.get("Time", None)
        if state.get("Thing", None) is not None:
            self.thing = frost_sta_client.model.thing.Thing()
            self.thing.__setstate__(state["Thing"])
        if state.get("Locations", None) is not None and isinstance(state["Locations"], list):
            entity_class = entity_type.EntityTypes['Location']['class']
            self.locations = utils.transform_json_to_entity_list(state['Locations'], entity_class)
            self.locations.next_link = state.get('Locations@iot.nextLink', None)

    def get_dao(self, service):
        return HistoricalLocationDao(service)



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import requests

from frost_sta_client.dao import *
from frost_sta_client.service import auth_handler


class SensorThingsService:

    def __init__(self, url, auth_handler=None):
        self.url = url
        self.auth_handler = auth_handler

    @property
    def auth_handler(self):
        return self._auth_handler

    @auth_handler.setter
    def auth_handler(self, value):
        if value is None:
            self._auth_handler = None
            return
        if not isinstance(value, auth_handler.AuthHandler):
            raise ValueError('auth should be of type AuthHandler!')
        self._auth_handler = value

    def execute(self, method, url, **kwargs):
        if self.auth_handler is not None:
            response = requests.request(method, url, auth=self.auth_handler.add_auth_header(), **kwargs)
        else:
            response = requests.request(method, url, **kwargs)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise e

        return response

    def create(self, entity):
        entity.get_dao(self).create(entity)

    def update(self, entity):
        entity.get_dao(self).update(entity)

    def patch(self, entity, patches):
        entity.get_dao(self).patch(entity, patches)

    def delete(self, entity):
        entity.get_dao(self).delete(entity)

    def actuators(self):
        return actuator.ActuatorDao(self)

    def datastreams(self):
        return datastream.DatastreamDao(self)

    def features_of_interest(self):
        return features_of_interest.FeaturesOfInterestDao(self)

    def historical_location(self):
        return historical_location.HistoricalLocationDao(self)

    def locations(self):
        return location.LocationDao(self)

    def multi_datastreams(self):
        return multi_datastream.MultiDatastreamDao(self)

    def observations(self):
        return observation.ObservationDao(self)

    def observed_properties(self):
        return observedproperty.ObservedPropertyDao(self)

    def sensors(self):
        return sensor.SensorDao(self)

    def tasks(self):
        return task.TaskDao(self)

    def tasking_capabilities(self):
        return tasking_capability.TaskingCapabilityDao(self)

    def things(self):
        return thing.ThingDao(self)



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import frost_sta_client.query.query
import frost_sta_client.utils

import logging
import requests
import jsonpatch
import json
from furl import furl


class BaseDao:
    """
    The entity independent implementation of a data access object. Specific entity Daos
    can be implemented by inheriting from this class.
    """
    APPLICATION_JSON_PATCH = {'Content-type': 'application/json-patch+json'}

    def __init__(self, service, entitytype):
        """
        Constructor.
        params:
            service: the service to operate on
            entitytype: a dictionary describing the type of the entity
        """
        self.service = service
        self.entitytype = entitytype["singular"]
        self.entitytype_plural = entitytype["plural"]
        self.entity_class = entitytype["class"]

    @property
    def service(self):
        return self._service

    @service.setter
    def service(self, value):
        if value is None or isinstance(value, frost_sta_client.service.sensorthingsservice.SensorThingsService):
            self._service = value
            return
        raise ValueError('service should be of type SensorThingsService')

    @property
    def entitytype(self):
        return self._entitytype

    @entitytype.setter
    def entitytype(self, value):
        if value is None or isinstance(value, str):
            self._entitytype = value
            return
        raise ValueError('entitytype should be of type String')

    @property
    def entitytype_plural(self):
        return self._entitytype_plural

    @entitytype_plural.setter
    def entitytype_plural(self, value):
        if value is None or isinstance(value, str):
            self._entitytype_plural = value
            return
        raise ValueError('entitytype_plural should be of type String')

    @property
    def entity_class(self):
        return self._entity_class

    @entity_class.setter
    def entity_class(self, value):
        if value is None or isinstance(value, str):
            self._entity_class = value
            return
        raise ValueError('entity_class should be of type string')

    def create(self, entity):
        url = furl(self.service.url)
        url.path.add(self.entitytype_plural)
        logging.debug('Posting to ' + str(url.url))
        json_dict = frost_sta_client.utils.transform_entity_to_json_dict(entity)
        try:
            response = self.service.execute('post', url, json=json_dict)
        except requests.exceptions.HTTPError as e:
            error_json = e.response.json()
            error_message = error_json['message']
            logging.error("Creating {} failed with status-code {}, {}".format(type(entity).__name__,
                                                                            e.response.status_code,
                                                                            error_message))
            raise e
        entity.id = int(frost_sta_client.utils.extract_value(response.headers['location']))
        entity.service = self.service
        logging.debug('Received response: ' + str(response.status_code))

    def patch(self, entity, patches):
        """
        method to patch STA entities
        param entity: entity, that the patches should be applied to
        param patches: either a JsonPatch object or list of dictionaries, containing jsonpatch commands
        """
        url = furl(self.service.url)
        if entity.id is None or entity.id == '':
            raise AttributeError('please provide an entity with a valid id')
        url.path.add(self.entity_path(entity.id))
        logging.debug(f'Patching to {url.url}')
        headers = self.APPLICATION_JSON_PATCH
        if patches is None:
            raise ValueError('please provide a list of patches, either as a jsonpatch object or a '
                             'list of dictionaries')
        if not isinstance(patches, jsonpatch.JsonPatch) and \
                not (isinstance(patches, list) and all(isinstance(x, dict) for x in patches)):
            raise ValueError('please provide a list of patches, either as a jsonpatch object or a '
                             'list of dictionaries')
        if isinstance(patches, jsonpatch.JsonPatch):
            patches = patches.patch
        try:
            response = self.service.execute('patch', url, json=patches, headers=headers)
        except requests.exceptions.HTTPError as e:
            error_json = e.response.json()
            error_message = error_json['message']
            logging.error("Patching {} failed with status-code {}, {}".format(type(entity).__name__,
                                                                            e.response.status_code,
                                                                            error_message))
            raise e
        logging.debug(f'Received response: {str(response.status_code)}')

    def update(self, entity):
        url = furl(self.service.url)
        if entity.id is None or entity.id == '':
            raise AttributeError('please provide an entity with a valid id')
        url.path.add(self.entity_path(entity.id))
        logging.debug('Updating to {}'.format(url.url))
        json_dict = frost_sta_client.utils.transform_entity_to_json_dict(entity)
        try:
            response = self.service.execute('put', url, json=json_dict)
        except requests.exceptions.HTTPError as e:
            error_json = e.response.json()
            error_message = error_json['message']
            logging.error("Updating {} failed with status-code {}, {}".format(type(entity).__name__,
                                                                            e.response.status_code,
                                                                            error_message))
            raise e
        logging.debug('Received response: {}'.format(str(response.status_code)))

    def find(self, id):
        url = furl(self.service.url)
        url.path.add(self.entity_path(id))
        logging.debug('Fetching: {}'.format(url.url))
        try:
            response = self.service.execute('get', url)
        except requests.exceptions.HTTPError as e:
            error_json = e.response.json()
            error_message = error_json['message']
            logging.error("Finding {} failed with status-code {}, {}".format(id,
                                                                            e.response.status_code,
                                                                            error_message))
            raise e
        logging.debug('Received response: {}'.format(response.status_code))
        json_response = response.json()
        json_response['id'] = json_response['@iot.id']
        entity = frost_sta_client.utils.transform_json_to_entity(json_response, self.entity_class)
        return entity

    def delete(self, entity):
        url = furl(self.service.url)
        url.path.add(self.entity_path(entity.id))
        logging.debug('Deleting: {}'.format(url.url))
        try:
            response = self.service.execute('delete', url)
        except requests.exceptions.HTTPError as e:
            error_json = e.response.json()
            error_message = error_json['message']
            logging.error("Deleting {} failed with status-code {}, {}".format(type(entity).__name__,
                                                                            e.response.status_code,
                                                                            error_message))
            raise e
        logging.debug('Received response: {}'.format(response.status_code))

    def entity_path(self, id):
        return "{}({})".format(self.entitytype_plural, id)

    def query(self):
        return frost_sta_client.query.query.Query(self.service, self.entitytype, self.entitytype_plural, self.entity_class)



# Copyright (C) 2021 Fraunhofer Institut IOSB, Fraunhoferstr. 1, D 76131
# Karlsruhe, Germany.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import frost_sta_client.model.task
from . import entity
from . import thing
from . import actuator

from frost_sta_client import utils
from .ext import entity_type
from .ext import entity_list

from frost_sta_client.dao.tasking_capability import TaskingCapabilityDao


class TaskingCapability(entity.Entity):

    def __init__(self,
                 name='',
                 description='',
                 properties=None,
                 tasking_parameters=None,
                 tasks=None,
                 thing=None,
                 actuator=None,
                 **kwargs):
        super().__init__(**kwargs)
        if tasking_parameters is None:
            tasking_parameters = {}
        if properties is None:
            properties = {}
        self.name = name
        self.description = description
        self.tasking_parameters = tasking_parameters
        self.properties = properties
        self.tasks = tasks
        self.thing = thing
        self.actuator = actuator

    def __new__(cls, *args, **kwargs):
        new_tc = super().__new__(cls)
        attributes = {'_id': None, '_name': '', '_description': '', '_properties': {}, '_tasking_parameters': {},
                      '_tasks': None, '_thing': None, '_actuator': None, '_self_link': '', '_service': None}
        for key, value in attributes.items():
            new_tc.__dict__[key] = value
        return new_tc

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise ValueError('name should be of type str!')
        self._name = value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            raise ValueError('description should be of type str!')
        self._description = value

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        if not isinstance(value, dict):
            raise ValueError('properties should be of type dict!')
        self._properties = value

    @property
    def tasking_parameters(self):
        return self._tasking_parameters

    @tasking_parameters.setter
    def tasking_parameters(self, value):
        if value is None:
            self._tasking_parameters = {}
            return
        if not isinstance(value, dict):
            raise ValueError('Tasking parameters should be of type dict!')
        self._tasking_parameters = value

    @property
    def tasks(self):
        return self._tasks

    @tasks.setter
    def tasks(self, value):
        if value is None:
            self._tasks = None
            return
        if not isinstance(value, list) and all(isinstance(t, frost_sta_client.model.task.Task) for t in value):
            entity_class = entity_type.EntityTypes['Task']['class']
            self._tasks = entity_list.EntityList(entity_class=entity_class, entities=value)
            return
        if isinstance(value, entity_list.EntityList) \
                and all(isinstance(t, frost_sta_client.model.task.Task) for t in value.entities):
            self._tasks = value
            return
        raise ValueError('tasks should be of type Task!')

    @property
    def thing(self):
        return self._thing

    @thing.setter
    def thing(self, value):
        if value is None:
            self._thing = None
            return
        if not isinstance(value, thing.Thing):
            raise ValueError('thing should be of type Thing!')
        self._thing = value

    @property
    def actuator(self):
        return self._actuator

    @actuator.setter
    def actuator(self, value):
        if value is None:
            self._actuator = None
            return
        if not isinstance(value, actuator.Actuator):
            raise ValueError('actuator should be of type Actuator!')
        self._actuator = value

    def ensure_service_on_children(self, service):
        if self.actuator is not None:
            self.actuator.set_service(service)
        if self.thing is not None:
            self.thing.set_service(service)
        if self.tasks is not None:
            self.thing.set_service(service)

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, type(self)):
            return False
        if id(self) == id(other):
            return True
        if self.name != other.name:
            return False
        if self.description != other.description:
            return False
        if self.tasking_parameters != other.tasking_parameters:
            return False
        if self.properties != other.properties:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    def __getstate__(self):
        data = super().__getstate__()
        if data.get('@iot.id', None) is not None:
            return data
        if self.name is not None and self.name != '':
            data['name'] = self.name
        if self.description is not None and self.description != '':
            data['description'] = self.description
        if self.tasking_parameters is not None and self.tasking_parameters != {}:
            data['taskingParameters'] = self.tasking_parameters
        if self.properties is not None and self.properties != {}:
            data['properties'] = self.properties
        if self.thing is not None:
            data['Thing'] = self.thing
        if self.tasks is not None and len(self.tasks.entities) > 0:
            data['Tasks'] = self.tasks.__getstate__()
        if self.actuator is not None:
            data['Actuator'] = self.actuator
        return data

    def __setstate__(self, state):
        super().__setstate__(state)
        self.name = state.get('name', '')
        self.description = state.get('description', '')
        self.tasking_parameters = state.get('taskingParameters', {})
        self.properties = state.get('properties', {})
        if state.get('Tasks', None) is not None:
            entity_class = entity_type.EntityTypes['Task']['class']
            self.tasks = utils.transform_json_to_entity_list(state['Tasks'], entity_class)
            self.tasks.next_link = state.get('Tasks@iot.nextLink', None)
        if state.get('Actuator', None) is not None:
            self.actuator = frost_sta_client.model.actuator.Actuator()
            self.actuator.__setstate__(state['Actuator'])
        if state.get('Thing', None) is not None:
            self.thing = frost_sta_client.model.thing.Thing()
            self.thing.__setstate__(state['Thing'])

    def get_dao(self, service):
        return TaskingCapabilityDao(service)
